"""
语料处理相关常用函数
"""
import doctest
import pathlib  # py3


def load_stopwords(filepath, encoding="utf8"):
    """
    Args:
        filepath(str):
        encoding:

    Examples:
        >>> stopwords_en = load_stopwords("./stopwords/stopwords_en")
        >>> 'the' in stopwords_en
        True

    Returns:
        set
    """
    with open(filepath, encoding=encoding) as f:
        # 去空白符，转小写（不影响中文）
        stopwords = set(word.strip().lower() for word in f if not word.isspace())

    return stopwords


def remove_stopwords(tokens, stopwords, encoding='utf8'):
    """remove stopwords

    Args:
        tokens(list of str): a list of tokens/words
        stopwords(list of str or str): a list of words or a filepath
        encoding:

    Examples:
        >>> remove_stopwords(['huay', 'the'], ['the'])
        ['huay']

        >>> remove_stopwords(['huay', 'the'], 'stopwords_en')
        ['huay']

    """
    if isinstance(stopwords, str):
        stopwords = load_stopwords(stopwords, encoding)
    else:
        stopwords = set(stopwords)

    return [w for w in tokens if w not in stopwords]


def clear_str_en(text):
    """clear the en text

    """


if __name__ == '__main__':
    """"""
    doctest.testmod()

    # with open('stopwords_en', encoding='utf8') as f:
    #     stopwords = set(word.strip() for word in f if not word.isspace())
    #     print(stopwords)



"""文件相关

"""
from __future__ import print_function

import os
import pathlib
from six.moves import urllib

# __all__ = [
#     'maybe_download',
#     'rename_replace_sep',
#     'rename_batch',
#     'get_dir_filenames'
# ]
# 需要的直接加到 __init__.py 中


def maybe_download(url, to_path='D:/Tmp', filename=None, expected_byte=None):
    """下载文件到指定目录

    Args:
        url (str): 文件下载路径
        to_path (str): 下载到本地路径
        filename (str): 重命名文件
        expected_byte (int): 文件预期大小

    Returns:
        str: filepath

    Examples:

        >>> url = 'http://mattmahoney.net/dc/bbb.zip'
        >>> filepath = maybe_download(url, filename='b.zip')
        File is ready.
        >>> fp = maybe_download(url, to_path='D:/Tmp/b', expected_byte=45370)
        File is downloading.
        File has been damage, please download it manually.

    """
    if filename is not None:
        filepath = os.path.join(maybe_mkdirs(to_path), filename)
    else:
        _, filename = os.path.split(url)
        filepath = os.path.join(maybe_mkdirs(to_path), filename)

    if not os.path.exists(filepath):
        urllib.request.urlretrieve(url, filepath)
        print('File is downloading.')

        if expected_byte is not None:
            file_size = os.stat(filepath).st_size
            if file_size != expected_byte:
                print('File has been damage, please download it manually.')
    else:
        print('File is ready.')

    return filepath


def rename_replace_sep(filename, new_sep='_', prefix=None, lower_case=False, sep=None, maxsplit=-1):
    """重命名：替换分隔符

    Args:
        filename (str): 原文件名，不包含路径
        new_sep (str): 新分隔符，默认是 '_'
        prefix (str): 前缀，分隔符会自动添加
        lower_case (bool): 是否替换成小写
        sep (str): 需要替换的分隔符，默认是所有空白符
        maxsplit (int): 分割几段，-1 表示全部分割

    Returns:
        str

    Examples：

        >>> rename_replace_sep('a b c.txt', prefix='M')
        'M_a_b_c.txt'

    """
    if lower_case:
        filename.lower()
    if prefix is not None:
        filename = new_sep.join([prefix, filename])
    return new_sep.join(filename.split(sep, maxsplit))


def get_dir_filenames(dirpath, abspath=True, recursive=True, yield_=True):
    """获取目录下所有文件名 （默认递归）

    该函数仅用于需要一次性处理大量**同类文件**的情况

    又发现一个实用的库函数
        root, dirs, files = os.walk(path)
    该函数会递归遍历 path 下的所有文件夹

    注意：一旦函数体中出现 yield from 关键字，该函数就会被认为是一个生成器函数，
    无论该 yield from 语句是否被执行；
    因此，额外定义了两个函数 `_get_dir_filenames_list` 和 `_get_dir_filenames_yield`

    Args:
        dirpath (str): 文件夹路径
        abspath (bool): 是否返回绝对路径
        recursive (bool): 是否递归
        yield_ (bool): 是否使用生成器模式，实测确实可以"加速"

    Examples:

        >>> fs_list = get_dir_filenames('D:/Tmp', yield_=False)
        >>> fs_gen = get_dir_filenames('D:/Tmp')

        %timeit hy.get_dir_filenames('D:/Tmp', yield_=False)
        182 ms ± 1.13 ms per loop (mean ± std. dev. of 7 runs, 10 loops each)

        %timeit hy.get_dir_filenames('D:/Tmp')
        391 ns ± 1.69 ns per loop (mean ± std. dev. of 7 runs, 1000000 loops each)

    """
    if yield_:
        return _get_dir_filenames_yield(dirpath, abspath=abspath, recursive=recursive)
    else:
        return _get_dir_filenames_list(dirpath, abspath=abspath, recursive=recursive)


def _get_dir_filenames_list(dirpath, abspath=True, recursive=True):
    """获取目录下所有文件名 （返回列表）

    参数含义同 `get_dir_filenames()`

    """
    fs = []
    if recursive:
        for root, _, files in os.walk(dirpath):
            if abspath:
                fs.extend((os.path.join(root, file) for file in files))
            else:
                fs.extend(files)
    else:
        if abspath:
            fs.extend((os.path.join(dirpath, file) for file in os.listdir(dirpath)))
        else:
            fs.extend(os.listdir(dirpath))

    return fs


def _get_dir_filenames_yield(dirpath, abspath=True, recursive=True):
    """get_dir_filenames() 的生成器版本

    参数含义同 `get_dir_filenames()`

    """
    # if recursive:
    #     for root, _, files in os.walk(dirpath):
    #         if abspath:
    #             yield from (os.path.join(root, file) for file in files)
    #         else:
    #             yield from (file for file in files)
    # else:
    #     if abspath:
    #         yield from (os.path.join(dirpath, file) for file in os.listdir(dirpath))
    #     else:
    #         yield from (file for file in os.listdir(dirpath))

    yield from _get_dir_filenames_list(dirpath, abspath=abspath, recursive=recursive)


def rename_batch(dirpath, rename_fn=rename_replace_sep, recursive=True):
    """文件批量重命名

    默认行为是将文件名中的空白符替换成 '_'

    Args:
        dirpath (str): 文件目录
        rename_fn (callable): 重命名方法
        recursive (bool): 是否递归

    Examples:

        >>> rename_batch('D:\Tmp')

    """
    fs = get_dir_filenames(dirpath, recursive=recursive)
    for f in fs:
        path, filename = os.path.split(f)
        new = rename_fn(filename)
        os.rename(f, os.path.join(path, new))


def rename_del_ext(filename):
    """重命名：删除后缀"""
    filename, _ = os.path.splitext(filename)
    return filename


def path_is_exist(path, is_file=False):
    """
    判断文件或文件夹是否存在

    Args:
        path(str): 文件或文件夹的路径

    Returns:

    """
    path = pathlib.Path(path)

    if is_file:
        return path.is_file()
    else:
        return path.exists()


def maybe_mkdirs(path, is_file=False, exist_ok=True):
    """（可能需要）创建文件夹，不会创建文件

    如果 path 是一个文件夹路径，则直接创建
    如果 path 是一个包含文件名的路径，则忽略文件名，创建路径
        注意，不能使用 pathlib，因为可能该文件是不存在的

    Args:
        path (str): 待创建的目录路径，递归创建
        filepath (bool): 如果 path 是一个文件路径，创建该文件依赖的目录
            设置该参数的目的主要是为了判断无后缀的文件，
            对于带后缀的文件会自动判断
        exist_ok (bool): 默认为 True

    Returns:
        str

    Examples:

        >>> maybe_mkdirs('D:/Tmp/a/b/')
        'D:/Tmp/a/b/'
        >>> maybe_mkdirs('D:/Tmp/a/b/c.txt')
        'D:/Tmp/a/b/c.txt'
        >>> maybe_mkdirs('D:/Tmp/a/b/c', is_file=True)  # c 是一个无后缀文件
        'D:/Tmp/a/b/c'

    """
    if is_file:
        dirs, filename = os.path.split(path)
        os.makedirs(dirs, exist_ok=exist_ok)
        return path

    if os.path.splitext(path)[1].startswith('.'):  # 是一个文件路径
        dirs, filename = os.path.split(path)
        os.makedirs(dirs, exist_ok=exist_ok)
    else:
        os.makedirs(path, exist_ok=exist_ok)

    return path



# -*- coding: utf-8 -*-
"""Example Google style docstrings.

This module demonstrates documentation as specified by the `Google Python Style Guide`_.
Sections are created with a section header and a colon followed by a block of indented text.

Examples:
    Examples can be given using either the ``Example`` or ``Examples`` sections.
    Sections support any reStructuredText formatting, including literal blocks::

        $ python example_google.py

Section breaks are created by resuming unindented text.
Section breaks are also implicitly created anytime a new section starts.

Attributes:
    module_level_variable_1 (int): Module level variables may be documented in
        either the ``Attributes`` section of the module docstring,
        or in an **inline docstring** immediately following the variable, like module_level_variable_2.

        Either form is acceptable, but the two should not be mixed.
        Choose one convention to document module level variables and be consistent with it.

Todo:
    * For module TODOs
    * next

.. _Google Python Style Guide:
    http://sphinxcontrib-napoleon.readthedocs.io/en/latest/example_google.html#example-google

"""

module_level_variable_1 = 12345

module_level_variable_2 = 98765
"""int: Module level variable documented inline.

The docstring may span multiple lines. The type may optionally be specified
on the first line, separated by a colon.
"""


def function_with_types_in_docstring(param1, param2):
    """Example function with types documented in the docstring.

    Args:
        param1 (int): The first parameter.
        param2 (str): The second parameter.

    Returns:
        bool: The return value. True for success, False otherwise.

    """


def function_with_pep484_type_annotations(param1: int, param2: str) -> bool:
    """Example function with PEP 484 type annotations.

    `PEP 484`_ type annotations are supported. If attribute, parameter, and
    return types are annotated according to `PEP 484`_, they do not need to be
    included in the docstring:

    Args:
        param1: The first parameter.
        param2: The second parameter.

    Returns:
        The return value. True for success, False otherwise.

    .. _PEP 484:
        https://www.python.org/dev/peps/pep-0484/

    """


def module_level_function(param1, param2=None, *args, **kwargs):
    """This is an example of a module level function.

    Function parameters should be documented in the ``Args`` section.
    The name of each parameter is required. The type and description of each parameter is optional.

    If \*args or \*\*kwargs are accepted,
    they should be listed as ``*args`` and ``**kwargs``.

    The format for a parameter is::

        name (type): description
            The description may span multiple lines.
            Following lines should be indented. The "(type)" is optional.

            Multiple paragraphs are supported in parameter descriptions.

    Args:
        param1 (int): The first parameter.
        param2 (str): The second parameter. Defaults to None.
            Second line of description should be indented.
        *args: Variable length argument list.
        **kwargs: Arbitrary keyword arguments.

    Returns:
        bool: True if successful, False otherwise.

        The return type is optional and may be specified at the beginning of
        the ``Returns`` section followed by a colon.

        The ``Returns`` section may span multiple lines and paragraphs.

        The ``Returns`` section supports any reStructuredText formatting,
        including literal blocks::

            {
                'param1': param1,
                'param2': param2
            }

    Examples:
        Examples should be written in doctest format, and should illustrate how to use the function.
        The Second should show the result, otherwise will failure in doctest

        >>> print(module_level_function(1, '2'))
        True

    Raises:
        AttributeError: The ``Raises`` section is a list of all exceptions
            that are relevant to the interface.
        ValueError: If `param2` is equal to `param1`.

    """
    if param1 == int(param2):
        raise ValueError('param1 may not be equal to param2')
    return True


def example_generator(n):
    """Generators have a ``Yields`` section instead of a ``Returns`` section.

    Args:
        n (list of int): a list of int

    Yields:
        int: The next number in the range of 0 to `n` - 1.

    """
    for i in n:
        yield i


class ExampleError(Exception):
    """Exceptions are documented in the same way as classes.

    The __init__ method may be documented in either the class level docstring,
    or as a docstring on the __init__ method itself.

    Either form is acceptable, but the two should not be mixed.
    Choose one convention to document the __init__ method and be consistent with it.

    Note:
        Do not include the `self` parameter in the ``Args`` section.

    Args:
        msg (str): Human readable string describing the exception.
        code (:obj:`int`, optional): Error code.

    Attributes:
        msg (str): Human readable string describing the exception.
        code (int): Exception error code.

    """

    def __init__(self, msg, code):
        self.msg = msg
        self.code = code


class ExampleClass(object):
    """The summary line for a class docstring should fit on one line.

    If the class has public attributes, they may be documented here
    in an ``Attributes`` section and follow the same formatting as a
    function's ``Args`` section.

    Alternatively, attributes may be documented inline with the attribute's declaration
    (see __init__ method below).

    Properties created with the ``@property`` decorator should be documented
    in the property's getter method.

    Attributes:
        attr1 (str): Description of `attr1`.
        attr2 (int): Description of `attr2`.

    """

    def __init__(self, param1, param2, param3):
        """Example of docstring on the __init__ method.

        The __init__ method may be documented in either the class level docstring,
        or as a docstring on the __init__ method itself.

        Either form is acceptable, but the two should not be mixed.
        Choose one convention to document the __init__ method and be consistent with it.

        Note:
            Do not include the `self` parameter in the ``Args`` section.

        Args:
            param1 (str): Description of `param1`.
            param2 (int): Description of `param2`.
                Multiple lines are supported.
            param3 (list of str): Description of `param3`.

        """
        self.attr1 = param1
        self.attr2 = param2
        self.attr3 = param3  #: Doc comment *inline* with attribute

        # list of str: Doc comment *before* attribute, with type specified
        self.attr4 = ['attr4']

        self.attr5 = None
        """str: Docstring *after* attribute, with type specified."""

    @property
    def readonly_property(self):
        """str: Properties should be documented in their getter method."""
        return 'readonly_property'

    @property
    def readwrite_property(self):
        """list of str: Properties with both a getter and setter
        should only be documented in their getter method.

        If the setter method contains notable behavior, it should be
        mentioned here.
        """
        return self.attr1

    @readwrite_property.setter
    def readwrite_property(self, value):
        self.attr1 = value

    def example_method(self, param1, param2):
        """Class methods are similar to regular functions.

        Note:
            Do not include the `self` parameter in the ``Args`` section.

        Args:
            param1: The first parameter.
            param2: The second parameter.

        Returns:
            True if successful, False otherwise.

        """
        return self.attr1 + param1 + param2

    def __special__(self):
        """By default special members with docstrings are not included.

        Special members are any methods or attributes that start with and
        end with a double underscore. Any special member with a docstring
        will be included in the output, if
        ``napoleon_include_special_with_doc`` is set to True.

        This behavior can be enabled by changing the following setting in
        Sphinx's conf.py::

            napoleon_include_special_with_doc = True

        """
        pass

    def __special_without_docstring__(self):
        pass

    def _private(self):
        """By default private members are not included.

        Private members are any methods or attributes that start with an
        underscore and are *not* special. By default they are not included in the output.

        This behavior can be changed such that private members *are* included
        by changing the following setting in Sphinx's conf.py::

            napoleon_include_private_with_doc = True

        """
        pass

    def _private_without_docstring(self):
        pass



from __future__ import print_function

import unittest
from unittest import TestCase

from huaytools.template.algorithm.bin_search import BinSearch


class TestBinSearch(TestCase):
    def setUp(self):
        self.bs = BinSearch()

    def test_lower_bound(self):
        xs = [1, 2, 3]
        target = 2
        ex_ret = 1
        self.assertEqual(self.bs.lower_bound(xs, target), ex_ret)

        xs = [1, 2, 2, 2, 3]
        target = 2
        ex_ret = 1
        self.assertEqual(self.bs.lower_bound(xs, target), ex_ret)

        xs = [1, 2, 3]
        target = 0
        ex_ret = 0
        self.assertEqual(self.bs.lower_bound(xs, target), ex_ret)

        xs = [1, 2, 3]
        target = 5
        ex_ret = 3
        self.assertEqual(self.bs.lower_bound(xs, target), ex_ret)

    def test_upper_bound(self):
        xs = [1, 2, 3]
        target = 2
        ex_ret = 2
        self.assertEqual(self.bs.upper_bound(xs, target), ex_ret)

        xs = [1, 2, 2, 2, 3]
        target = 2
        ex_ret = 4
        self.assertEqual(self.bs.upper_bound(xs, target), ex_ret)

        xs = [1, 2, 3]
        target = 0
        ex_ret = 0
        self.assertEqual(self.bs.upper_bound(xs, target), ex_ret)

        xs = [1, 2, 3]
        target = 5
        ex_ret = 3
        self.assertEqual(self.bs.upper_bound(xs, target), ex_ret)

    def test_bin_search(self):
        xs = [1, 2, 3]
        target = 2
        ex_ret = 1
        self.assertEqual(self.bs.lower_bound(xs, target), ex_ret)

        xs = [1, 2, 2, 2, 3]
        target = 2
        ex_ret = 1
        self.assertEqual(self.bs.lower_bound(xs, target), ex_ret)

        xs = [1, 2, 3]
        target = 0
        ex_ret = 0
        self.assertEqual(self.bs.lower_bound(xs, target), ex_ret)

        xs = [1, 2, 3]
        target = 5
        ex_ret = 3
        self.assertEqual(self.bs.lower_bound(xs, target), ex_ret)

    def test_bin_search_float(self):
        lb, ub = 3, 4
        ex_ret = 3.141
        self.assertEqual(self.bs.bin_search_float(lb, ub), ex_ret)

        lb, ub = 3, 4
        ex_ret = 3.1401
        self.assertEqual(self.bs.bin_search_float(lb, ub, precision=0.0001), ex_ret)

        lb, ub = -1, 10000
        ex_ret = 3.1401
        self.assertEqual(self.bs.bin_search_float(lb, ub, precision=0.0001), ex_ret)

    def test_bin_search_float_while(self):
        lb, ub = 3, 4
        ex_ret = 3.141
        self.assertEqual(self.bs.bin_search_float_while(lb, ub), ex_ret)

        lb, ub = 3, 4
        ex_ret = 3.1401
        self.assertEqual(self.bs.bin_search_float_while(lb, ub, precision=0.0001), ex_ret)

        lb, ub = -1, 10000
        ex_ret = 3.1401
        self.assertEqual(self.bs.bin_search_float_while(lb, ub, precision=0.0001), ex_ret)


if __name__ == '__main__':
    unittest.main()



"""
Apply the bunch into Python3, the original lib ref `pip install bunch`

Bunch is a subclass of dict with attribute-style access.

    >>> b = Bunch()
    >>> b.hello = 'world'
    >>> b.hello
    'world'
    >>> b['hello'] += "!"
    >>> b.hello
    'world!'
    >>> b.foo = Bunch(lol=True)
    >>> b.foo.lol
    True
    >>> b.foo is b['foo']
    True

    It is safe to import * from this module:

        __all__ = ('Bunch', 'bunchify','unbunchify')

    un/bunchify provide dictionary conversion; Bunches can also be
    converted via Bunch.to/fromDict().
"""

__version__ = '1.0.1'
VERSION = tuple(map(int, __version__.split('.')))

__all__ = ('Bunch', 'bunchify', 'unbunchify',)


class Bunch(dict):
    """ A dictionary that provides attribute-style access.

        >>> b = Bunch()
        >>> b.hello = 'world'
        >>> b.hello
        'world'
        >>> b['hello'] += "!"
        >>> b.hello
        'world!'
        >>> b.foo = Bunch(lol=True)
        >>> b.foo.lol
        True
        >>> b.foo is b['foo']
        True

        A Bunch is a subclass of dict; it supports all the methods a dict does...

        >>> b.keys()
        dict_keys(['hello', 'foo'])

        Including update()...

        >>> b.update({ 'ponies': 'are pretty!' }, hello=42)
        >>> print(repr(b))
        Bunch(foo=Bunch(lol=True), hello=42, ponies='are pretty!')

        As well as iteration...

        >>> [ (k, b[k]) for k in b ]
        [('hello', 42), ('foo', Bunch(lol=True)), ('ponies', 'are pretty!')]

        And "splats".

        >>> "The {knights} who say {ni}!".format(**Bunch(knights='lolcats', ni='can haz'))
        'The lolcats who say can haz!'

        See unbunchify/Bunch.toDict, bunchify/Bunch.fromDict for notes about conversion.
    """

    def __contains__(self, k):
        """
            >>> b = Bunch(ponies='are pretty!')
            >>> 'ponies' in b
            True
            >>> 'foo' in b
            False
            >>> b['foo'] = 42
            >>> 'foo' in b
            True
            >>> b.hello = 'hai'
            >>> 'hello' in b
            True
        """
        try:
            return hasattr(self, k) or dict.__contains__(self, k)
        except:
            return False

    # only called if k not found in normal places
    def __getattr__(self, k):
        """ Gets key if it exists, otherwise throws AttributeError.

            nb. __getattr__ is only called if key is not found in normal places.

            >>> b = Bunch(bar='baz', lol={})
            >>> b.foo
            Traceback (most recent call last):
                ...
            AttributeError: foo

            >>> b.bar
            'baz'
            >>> getattr(b, 'bar')
            'baz'
            >>> b['bar']
            'baz'

            >>> b.lol is b['lol']
            True
            >>> b.lol is getattr(b, 'lol')
            True
        """
        try:
            # Throws exception if not in prototype chain
            return object.__getattribute__(self, k)
        except AttributeError:
            try:
                return self[k]
            except KeyError:
                raise AttributeError(k)

    def __setattr__(self, k, v):
        """ Sets attribute k if it exists, otherwise sets key k. A KeyError
            raised by set-item (only likely if you subclass Bunch) will
            propagate as an AttributeError instead.

            >>> b = Bunch(foo='bar', this_is='useful when subclassing')
            >>> b.values                            #doctest: +ELLIPSIS
            <built-in method values of Bunch object at 0x...>
            >>> b.values = 'uh oh'
            >>> b.values
            'uh oh'
            >>> b['values']
            Traceback (most recent call last):
                ...
            KeyError: 'values'
        """
        try:
            # Throws exception if not in prototype chain
            object.__getattribute__(self, k)
        except AttributeError:
            try:
                self[k] = v
            except:
                raise AttributeError(k)
        else:
            object.__setattr__(self, k, v)

    def __delattr__(self, k):
        """ Deletes attribute k if it exists, otherwise deletes key k. A KeyError
            raised by deleting the key--such as when the key is missing--will
            propagate as an AttributeError instead.

            >>> b = Bunch(lol=42)
            >>> del b.lol
            >>> 'lol' in b
            False
        """
        try:
            # Throws exception if not in prototype chain
            object.__getattribute__(self, k)
        except AttributeError:
            try:
                del self[k]
            except KeyError:
                raise AttributeError(k)
        else:
            object.__delattr__(self, k)

    def toDict(self):
        """ Recursively converts a bunch back into a dictionary.

            >>> b = Bunch(foo=Bunch(lol=True), hello=42, ponies='are pretty!')
            >>> b.toDict()
            {'foo': {'lol': True}, 'hello': 42, 'ponies': 'are pretty!'}

            See unbunchify for more info.
        """
        return unbunchify(self)

    def __repr__(self):
        """ Invertible* string-form of a Bunch.

            >>> b = Bunch(foo=Bunch(lol=True), hello=42, ponies='are pretty!')
            >>> print(repr(b))
            Bunch(foo=Bunch(lol=True), hello=42, ponies='are pretty!')
            >>> eval(repr(b))
            Bunch(foo=Bunch(lol=True), hello=42, ponies='are pretty!')

            (*) Invertible so long as collection contents are each repr-invertible.
        """
        keys = self.keys()
        # keys.sort()
        keys = sorted(keys)
        args = ', '.join(['%s=%r' % (key, self[key]) for key in keys])
        return '%s(%s)' % (self.__class__.__name__, args)

    @staticmethod
    def fromDict(d):
        """ Recursively transforms a dictionary into a Bunch via copy.

            >>> b = Bunch.fromDict({'urmom': {'sez': {'what': 'what'}}})
            >>> b.urmom.sez.what
            'what'

            See bunchify for more info.
        """
        return bunchify(d)


# While we could convert abstract types like Mapping or Iterable, I think
# bunchify is more likely to "do what you mean" if it is conservative about
# casting (ex: isinstance(str,Iterable) == True ).
#
# Should you disagree, it is not difficult to duplicate this function with
# more aggressive coercion to suit your own purposes.

def bunchify(x):
    """ Recursively transforms a dictionary into a Bunch via copy.

        >>> b = bunchify({'urmom': {'sez': {'what': 'what'}}})
        >>> b.urmom.sez.what
        'what'

        bunchify can handle intermediary dicts, lists and tuples (as well as
        their subclasses), but ymmv on custom datatypes.

        >>> b = bunchify({ 'lol': ('cats', {'hah':'i win again'}),
        ...         'hello': [{'french':'salut', 'german':'hallo'}] })
        >>> b.hello[0].french
        'salut'
        >>> b.lol[1].hah
        'i win again'

        nb. As dicts are not hashable, they cannot be nested in sets/frozensets.
    """
    if isinstance(x, dict):
        return Bunch((k, bunchify(v)) for k, v in x.items())
    elif isinstance(x, (list, tuple)):
        return type(x)(bunchify(v) for v in x)
    else:
        return x


def unbunchify(x):
    """ Recursively converts a Bunch into a dictionary.

        >>> b = Bunch(foo=Bunch(lol=True), hello=42, ponies='are pretty!')
        >>> unbunchify(b)
        {'foo': {'lol': True}, 'hello': 42, 'ponies': 'are pretty!'}

        unbunchify will handle intermediary dicts, lists and tuples (as well as
        their subclasses), but ymmv on custom datatypes.

        >>> b = Bunch(foo=['bar', Bunch(lol=True)], hello=42,
        ...         ponies=('are pretty!', Bunch(lies='are trouble!')))
        >>> unbunchify(b) #doctest: +NORMALIZE_WHITESPACE
        {'foo': ['bar', {'lol': True}], 'hello': 42, 'ponies': ('are pretty!', {'lies': 'are trouble!'})}

        nb. As dicts are not hashable, they cannot be nested in sets/frozensets.
    """
    if isinstance(x, dict):
        return dict((k, unbunchify(v)) for k, v in x.items())
    elif isinstance(x, (list, tuple)):
        return type(x)(unbunchify(v) for v in x)
    else:
        return x


### Serialization

try:
    try:
        import json
    except ImportError:
        import simplejson as json


    def to_JSON(self, **options):
        """ Serializes this Bunch to JSON. Accepts the same keyword options as `json.dumps()`.

            >>> b = Bunch(foo=Bunch(lol=True), hello=42, ponies='are pretty!')
            >>> json.dumps(b)
            '{"foo": {"lol": true}, "hello": 42, "ponies": "are pretty!"}'
            >>> b.toJSON()
            '{"foo": {"lol": true}, "hello": 42, "ponies": "are pretty!"}'
        """
        return json.dumps(self, **options)


    Bunch.toJSON = to_JSON

except ImportError:
    pass

try:
    # Attempt to register ourself with PyYAML as a representer
    import yaml
    from yaml.representer import Representer, SafeRepresenter


    def from_yaml(loader, node):
        """ PyYAML support for Bunches using the tag `!bunch` and `!bunch.Bunch`.

            >>> yaml.load('''
            ... Flow style: !bunch.Bunch { Clark: Evans, Brian: Ingerson, Oren: Ben-Kiki }
            ... Block style: !bunch
            ...   Clark : Evans
            ...   Brian : Ingerson
            ...   Oren  : Ben-Kiki
            ... ''') #doctest: +NORMALIZE_WHITESPACE
            {'Flow style': Bunch(Brian='Ingerson', Clark='Evans', Oren='Ben-Kiki'),
             'Block style': Bunch(Brian='Ingerson', Clark='Evans', Oren='Ben-Kiki')}

            This module registers itself automatically to cover both Bunch and any
            subclasses. Should you want to customize the representation of a subclass,
            simply register it with PyYAML yourself.
        """
        data = Bunch()
        yield data
        value = loader.construct_mapping(node)
        data.update(value)


    def to_yaml_safe(dumper, data):
        """ Converts Bunch to a normal mapping node, making it appear as a
            dict in the YAML output.

            >>> b = Bunch(foo=['bar', Bunch(lol=True)], hello=42)
            >>> import yaml
            >>> yaml.safe_dump(b, default_flow_style=True)
            '{foo: [bar, {lol: true}], hello: 42}\\n'
        """
        return dumper.represent_dict(data)


    def to_yaml(dumper, data):
        """ Converts Bunch to a representation node.

            >>> b = Bunch(foo=['bar', Bunch(lol=True)], hello=42)
            >>> yaml.dump(b, default_flow_style=True)
            '!bunch.Bunch {foo: [bar, !bunch.Bunch {lol: true}], hello: 42}\\n'
        """
        return dumper.represent_mapping(u'!bunch.Bunch', data)


    yaml.add_constructor(u'!bunch', from_yaml)
    yaml.add_constructor(u'!bunch.Bunch', from_yaml)

    SafeRepresenter.add_representer(Bunch, to_yaml_safe)
    SafeRepresenter.add_multi_representer(Bunch, to_yaml_safe)

    Representer.add_representer(Bunch, to_yaml)
    Representer.add_multi_representer(Bunch, to_yaml)

    # Instance methods for YAML conversion
    def to_YAML(self, **options):
        """ Serializes this Bunch to YAML, using `yaml.safe_dump()` if
            no `Dumper` is provided. See the PyYAML documentation for more info.

            >>> b = Bunch(foo=['bar', Bunch(lol=True)], hello=42)
            >>> yaml.safe_dump(b, default_flow_style=True)
            '{foo: [bar, {lol: true}], hello: 42}\\n'
            >>> b.toYAML(default_flow_style=True)
            '{foo: [bar, {lol: true}], hello: 42}\\n'
            >>> yaml.dump(b, default_flow_style=True)
            '!bunch.Bunch {foo: [bar, !bunch.Bunch {lol: true}], hello: 42}\\n'
            >>> b.toYAML(Dumper=yaml.Dumper, default_flow_style=True)
            '!bunch.Bunch {foo: [bar, !bunch.Bunch {lol: true}], hello: 42}\\n'
        """
        opts = dict(indent=4, default_flow_style=False)
        opts.update(options)
        if 'Dumper' not in opts:
            return yaml.safe_dump(self, **opts)
        else:
            return yaml.dump(self, **opts)

    def from_YAML(*args, **kwargs):
        return bunchify(yaml.load(*args, **kwargs))


    Bunch.toYAML = Bunch.__str__ = to_YAML
    Bunch.fromYAML = staticmethod(from_YAML)

except ImportError:
    pass

if __name__ == "__main__":
    import doctest

    doctest.testmod()



import os

from urllib.request import urlopen
from bs4 import BeautifulSoup


def get_html(url, save=True, save_path='.', save_name='test.html'):
    """获取网页源码

    快速获取网页源码，并下载到本地文件，用于分析结构

    Args:
        url:
        save:
        save_path:
        save_name:

    Returns:

    """

    html_str = urlopen(url).read()
    soup = BeautifulSoup(html_str, "html.parser").prettify()

    if save:
        with open(os.path.join(save_path, save_name), 'w+', encoding="utf-8") as f:
            f.write(soup)

    return soup



"""工具函数（如果不知道应该归到哪个模块文件，就放在这里）
"""
import os
import sys
import platform
from typing import Iterable

import logging


if sys.version_info[0] >= 3:
    unicode = str


def yield_cycling(iterator):
    """
    无限循环迭代器

    itertools 提供了类似的方法
        from itertools import cycle

    Args:
        iterator (Iterable): 可迭代对象

    Examples:

        >>> it = yield_cycling([1, 2, 3])
        >>> for _ in range(4):
        ...     print(next(it))
        1
        2
        3
        1

    """
    while True:
        yield from iter(iterator)


def system_is(system_type):
    """
    判断系统类型

    Args:
        system_type(str): 系统类型，可选 "linux", "windows", ..

    Returns:
        bool

    Examples:
        >>> if system_is("windows"):
        ...     print("Windows")
        Windows

    """
    if system_type.lower().startswith('win'):
        system_type = 'Windows'
    else:
        system_type = 'Linux'
    return system_type == platform.system()


def system_is_windows():
    """
    If the system is windows, return True

    Examples:
        >>> if system_is_windows():
        ...     print("Windows")
        Windows
    """
    return system_is("Windows")


def get_logger(name=None, fname=None, mode='a', level=logging.INFO, stream=None,
               fmt="[%(name)s] : %(asctime)s : %(levelname)s : %(message)s"):
    """
    Default log to console. If `fname` is not None, if will log to the file at same time.

    Returns:


    Examples:
        >>> logger = get_logger(stream=sys.stdout, fmt="[%(name)s] : %(levelname)s : %(message)s")
        >>> logger.info("test")
        [tools.py] : INFO : test


    """
    # 虽然使用 root logger 可能会产生重复输出的问题，但为了能够同时输出不同的 handler，还是默认使用 root
    # if name is None:
    #     # name = os.path.splitext(os.path.basename(__file__))[0]
    #     name = os.path.basename(__file__)

    logger = logging.Logger(name)
    logger.setLevel(level)

    fmt = logging.Formatter(fmt)

    ch = logging.StreamHandler(stream)
    ch.setFormatter(fmt)

    logger.addHandler(ch)

    if fname is not None:
        fh = logging.FileHandler(fname, mode)
        fh.setFormatter(fmt)

        logger.addHandler(fh)

    return logger


def to_unicode(txt, encoding='utf8', errors='strict'):
    """Convert text to unicode.

    Args:
        txt:
        encoding:
        errors:

    Returns:
        str

    """

    if isinstance(txt, unicode):
        return txt
    return unicode(txt, encoding, errors=errors)


def get_var_name(**kwargs):
    """
    TODO(huay): It's useless.

    Get the literal name of var.
        s = 123
        it will return 's' rather than 123

    Examples:
        >>> s = 123
        >>> print(get_var_name(s=s))
        s

    Args:
        var:

    Returns:
        str

    """
    return list(kwargs.keys())[0]


def set_logging_basic_config(**kwargs):
    """
    Args can be specified:
        filename: Specifies that a FileHandler be created, using the specified
            filename, rather than a StreamHandler.
        filemode: Specifies the mode to open the file, if filename is specified
            (if filemode is unspecified, it defaults to 'a').
        format: Use the specified format string for the handler.
        datefmt: Use the specified date/time format.
        style: If a format string is specified, use this to specify the
            type of format string (possible values '%', '{', '$', for
            %-formatting, :meth:`str.format` and :class:`string.Template`
            - defaults to '%').
        level: Set the root logger level to the specified level.
        stream: Use the specified stream to initialize the StreamHandler. Note
            that this argument is incompatible with 'filename' - if both
            are present, 'stream' is ignored.
        handlers: If specified, this should be an iterable of already created
            handlers, which will be added to the root handler. Any handler
            in the list which does not have a formatter assigned will be
            assigned the formatter created in this function.

    Returns:
        None
    """
    if 'format' not in kwargs:
        kwargs['format'] = '[%(name)s] : %(asctime)s : %(levelname)s : %(message)s'
    if 'level' not in kwargs:
        kwargs['level'] = logging.INFO

    logging.basicConfig(**kwargs)



"""
Handle the different kinds of word embedding
    * word2vec
    * GloVe
    * FastText
"""
import doctest
import logging
from numpy import float32
from gensim.models import KeyedVectors

logger = logging.getLogger(__name__)


class BasicFormatWV:
    """
    The basic word vector format is the original C word2vec-tool format (text and binary).
    The first line is `vocab_size vector_size`, than are `token wordvec`. Which like
        ```
        2519370 300
        the -0.065334 -0.093031 -0.017571 ...
        of 0.048804 -0.28528 0.018557 ...
        ```
    """

    def __init__(self, fname, load_at_once=False):
        """

        """
        self.fname = fname
        self.word2vec = None
        self.word2id = dict()

        if load_at_once:
            self.load(fname)

    def load(self, fname=None, fvocab=None, binary=False, encoding='utf8', unicode_errors='strict', limit=None,
             datatype=float32):
        if fname is None:
            fname = self.fname

        self.word2vec = KeyedVectors.load_word2vec_format(fname,
                                                          fvocab=fvocab, binary=binary,
                                                          encoding=encoding, unicode_errors=unicode_errors,
                                                          limit=limit, datatype=datatype)
        return self

    def get_word2id_dict(self):
        """Build the word2id dict from the pre-train wordvec

        Examples:
            >>> wv_path = "D:/OneDrive/workspace/data/nlp/word2vec/fastText/wiki.en.vec"
            >>> basic_format_wv = BasicFormatWV(wv_path)
            >>> _ = basic_format_wv.load(limit=100)
            >>> word2id = basic_format_wv.get_word2id_dict()

        """
        if self.word2vec is None:
            logger.error("Load the wordvec first.")

        for k, v in self.word2vec.vocab.items():
            self.word2id[k] = v.index

        return self.word2id


class FastText:
    """
    The process to use FastText Word embedding
    """

    def load(self):
        """"""


if __name__ == '__main__':
    """"""
    doctest.testmod()



# modules
from . import utils
from . import crawler
from . import template
from . import nlp

# utils
from .utils.bunch import Bunch, bunchify, unbunchify

# utils.file_path
from .utils.file_path import rename_replace_sep
from .utils.file_path import rename_del_ext
from .utils.file_path import rename_batch
from .utils.file_path import maybe_download
from .utils.file_path import maybe_mkdirs
from .utils.file_path import get_dir_filenames

# utils.time
from .utils.time import timing
from .utils.time import str_to_date
from .utils.time import date_to_time
from .utils.time import get_timedelta

# utils.tools
from .utils.tools import yield_cycling
from .utils.tools import system_is
from .utils.tools import get_logger
from .utils.tools import set_logging_basic_config
from .utils.tools import to_unicode
from .utils.tools import get_var_name

# utils.save_load
from .utils.save_load import save_to_pickle
from .utils.save_load import load_from_pickle

# crawler
from .crawler.tools import get_html

# template
from .template import algorithm as algorithm



"""
stop words collections
"""
import os
import sys
from .corpus import load_stopwords

base_path = os.path.dirname(__file__)
sys.path.append(base_path)

stopwords_en = {'i', 'me', 'my', 'myself', 'we', 'our', 'ours', 'ourselves',
                'you', "you're", "you've", "you'll", "you'd", 'your', 'yours', 'yourself', 'yourselves',
                'he', 'him', 'his', 'himself',
                'she', "she's", 'her', 'hers', 'herself',
                'it', "it's", 'its', 'itself',
                'they', 'them', 'their', 'theirs', 'themselves',
                'what', 'which', 'who', 'whom',
                'this', 'that', "that'll", 'these', 'those',
                'am', 'is', 'are', 'was', 'were', 'be', 'been', 'being',
                'have', 'has', 'had', 'having',
                'do', 'does', 'did', 'doing',
                'a', 'an', 'the',
                'and', 'but', 'if', 'or', 'because', 'as', 'until',
                'while', 'of', 'at', 'by', 'for', 'with',
                'about', 'against', 'between', 'into', 'through', 'during', 'before',
                'after', 'above', 'below', 'to', 'from', 'up', 'down', 'in',
                'out', 'on', 'off', 'over', 'under', 'again', 'further', 'then', 'once',
                'here', 'there',
                'when', 'where', 'why', 'how',
                'all', 'any', 'both', 'each',
                'few', 'more', 'most', 'other', 'some', 'such',
                'no', 'nor', 'not', 'only', 'own', 'same',
                'so', 'than', 'too', 'very',
                's', 't',
                'can', 'will', 'just',
                'don', "don't", 'should', "should've",
                'now', 'd', 'll', 'm', 'o', 're', 've', 'y', 'ain',
                'aren', "aren't",
                'couldn', "couldn't",
                'didn', "didn't", 'doesn', "doesn't",
                'hadn', "hadn't", 'hasn',
                "hasn't", 'haven', "haven't",
                'isn', "isn't", 'ma',
                'mightn', "mightn't", 'mustn', "mustn't", 'needn', "needn't",
                'shan', "shan't", 'shouldn', "shouldn't",
                'wasn', "wasn't", 'weren', "weren't",
                'won', "won't", 'wouldn', "wouldn't"}
"""
The stopwords come from NLTK.

You can get the original file as bellow:
    ```
    nltk.download('stopwords')

    from nltk.corpus import stopwords
    stopwords_en = stopwords.words('english')

    ```
"""

stopwords_zh = load_stopwords(os.path.join(base_path, "stopwords", "stopwords_zh"))



"""搜索算法
"""
from __future__ import print_function
from __future__ import division

# import operator
import math


# from decimal import Decimal


class BinSearch:

    @staticmethod
    def bin_search_float(lb, ub, target=3.14, cond_fn=lambda x, y: x <= y, precision=0.001, n_steps=100):
        """二分查找求解模板，一般用于求浮点值解
        本例的默认行为是求*大于*3.14 的最小值，精确到 3 位小数 (3.141)
        迭代固定次数可以回避终止条件，代价是多余的循环，
        比如本例取 (3,4) 为边界的话，只要6次就能找到满足条件的解，具体参考 `bin_search_float_while()`

        Notes:
            处理精度问题时，注意是 round, math.ceil 还是 math.floor
            这里应该是 math.ceil；因为返回值大概是 3.1400000001 这种形式
            此外，round 并不是精确的四舍五入，具体百度

        Args:
            lb (float or int): 解的搜索上限
            ub (float or int): 解的搜索下限
            target (float):
            cond_fn (callable): 条件函数
            precision (float): 精确到 precision 位小数
            n_steps (int): 迭代次数，默认值为 100
                100 次循环可以达到 10^(−30) 的精度，基本上能满足一般要求

        """
        for _ in range(n_steps):

            mid = lb + (ub - lb) / 2  # python 中的 / 默认是浮点除法

            if cond_fn(mid, target):  # 包含等号，已转为闭区间
                lb = mid
            else:
                ub = mid

        # 因为 lb 已经时闭区间，所以应该用 ub；当然这也应该视情况而定
        times = 1 / precision
        return math.ceil(ub * times) / times

    @staticmethod
    def bin_search_float_while(lb, ub, target=3.14, cond_fn=lambda x, y: x <= y, precision=0.001):
        """二分查找求解模板，不固定循环次数

        Args:
            lb:
            ub:
            target:
            cond_fn:
            precision:

        Returns:

        """

        while True:
            mid = lb + (ub - lb) / 2

            if cond_fn(mid, target):
                lb = mid
            else:
                ub = mid

            if 0 < mid - target < precision:
                break

        times = 1 / precision
        return math.ceil(ub * times) / times

    @staticmethod
    def bin_search(xs, target, cond_fn=lambda x, y: x < y):
        """二分查找通用模板，一般用来寻找正整数值
        通过指定条件函数 cond_fn 来搜索所需的值
        本例使用*闭区间*和`cond_fn`实现 lower_bound 相同的效果

        Notes:
            * 本例中将 (lb, ub) 看作闭区间
            * 不要苛求固定的模板

        Args:
            xs:
            target:
            cond_fn (callable): 条件函数，返回值应该是 bool 类型
                这里默认是*小于函数*

        """
        lb, ub = 0, len(xs) - 1

        while lb < ub:  # 闭区间，不使用 lb + 1 < ub
            mid = lb + (ub - lb) // 2

            if cond_fn(xs[mid], target):
                lb = mid + 1  # 一般不包含*等号*的地方需要 + 1，缩小区间
            else:
                ub = mid

        return lb  # 因为是闭区间，返回值也不需要 lb + 1

    @staticmethod
    def lower_bound(xs, target):
        """二分搜索 lower bound 版
        行为同 C++ <algorithm> 中的 lower_bound
        在给定升序数组中返回*大于等于* target 的最小索引；
        如果 target 大于数组中的最大值，返回 len(xs)

        Note:
            * 将 (lb, ub) 当作开区间，避免 lb = mid+1 或者 ub = mid-1 这种无谓的优化
                相应的，把 (lb, ub) 初始化为 (-1, n) 看作是 (-∞, +∞)
            * 有的问题可能当作闭区间更简单，不要苛求固定的模板

        Args:
            xs (list[int]): 有序数组
            target (int): 待查找的值

        Returns:
            int: 目标值（可能）的索引

        """
        lb, ub = -1, len(xs)  # (-1, n) 看作是 (-∞, +∞)

        while lb + 1 < ub:  # 既然是开区间，那么至少要有一个元素
            mid = lb + (ub - lb) // 2

            if xs[mid] < target:
                lb = mid
            else:
                ub = mid  # 因为包含等号，所以实际上这里的下界已经是*闭区间 (lb, ub]*了

        # 因为 lb 始终是开区间，所以需要返回 lb + 1，
        # 此时区间 (lb, ub] 内只有一个元素，所以返回 ub 也是一样的
        return lb + 1

    @staticmethod
    def upper_bound(xs, target):
        """二分搜索 upper bound 版
        行为同 C++ <algorithm> 中的 upper_bound
        在给定升序数组中返回*大于* target 的最小索引；
        注意只返回*大于* target 的索引，即使数组中包含 target
        如果 target 小于数组中的最小值，返回 0

        Args:
            xs (list[int]): 有序数组
            target (int): 待查找的值

        Returns:
            int: 目标值（可能）的索引

        """
        lb, ub = -1, len(xs)

        while lb + 1 < ub:
            mid = lb + (ub - lb) // 2

            if xs[mid] <= target:
                lb = mid  # 因为包含等号，这里实际上已经是*闭区间 [lb, ub)*了
            else:
                ub = mid

        # 最终区间内只有一个元素 [lb, ub)
        # 因为要满足 upper_bound 大于 target 的行为，这里依然返回的是 lb + 1
        return lb + 1



"""
Common tool functions
"""
import re
import doctest

TOKENIZER_RE = re.compile(r"[A-Z]{2,}(?![a-z])|[A-Z][a-z]+(?=[A-Z])|[\'\w\-]+",
                          re.UNICODE)


def tokenizer(iterator, tokenizer_re=TOKENIZER_RE):
    """Tokenizer generator.

    Examples:
        >>> list(tokenizer(["'sss'", 'dsa2ds', '3$ dasd']))

    Args:
        iterator: Input iterator with strings.
        tokenizer_re:

    Yields:
        array of tokens per each value in the input.
    """
    for value in iterator:
        yield tokenizer_re.findall(value)


if __name__ == '__main__':
    doctest.testmod()



"""
保存到各种类型的文件
"""
import huaytools
import pickle


def save_to_pickle(obj, fname):
    """
    保存到 pickle 文件

    Args:
        obj: 需要保存的对象
        fname(str): 文件名

    Returns:
        None

    """
    fname = huaytools.maybe_mkdirs(fname, is_file=True)
    with open(fname, 'wb') as f:
        pickle.dump(obj, f)


def load_from_pickle(fname):
    """
    从 pickle 加载对象

    Args:
        fname(str): 文件名

    Returns:

    """
    with open(fname) as f:
        return pickle.load(f)



"""
* Project Name: huaytools
* Author: huay
* Mail: imhuay@163.com
* Created Time:  2018-1-26 11:33:13
"""
from setuptools import setup, find_packages

install_requires = [
    'six',
    'bs4',
    # 'bunch',  # add the hole bunch in, and apply it to Python3

    # Use Anaconda to install avoid install failed
    # 'tensorflow',  # 1.6
    # 'gensim',
    # 'numpy',
]

setup(
    name="huaytools",
    version="0.2.10",
    keywords=("huay", "huaytools"),
    description="huay's tools",
    long_description="huay's tools",
    license="MIT Licence",
    url="https://github.com/imhuay/huaytools",
    author="huay",
    author_email="imhuay@163.com",
    packages=find_packages(),
    include_package_data=True,
    platforms="any",
    install_requires=install_requires
)



import time
import functools

from datetime import datetime, timedelta


def timing(func):
    """函数计时装饰器

    Args:
        func: 被装饰的函数

    Examples:

        >>> @timing
        ... def t(seconds=2):
        ...     return time.sleep(seconds)
        ...
        >>> ret, duration = t()

    Returns:
        (Any, float): 返回一个二元组 (原函数的返回值, 函数用时)

    """

    @functools.wraps(func)
    def inner(*args, **kwargs):
        start = time.time()
        ret = func(*args, **kwargs)
        return ret, time.time() - start

    return inner


def str_to_date(date_str, fmt='%Y-%m-%d %H:%M'):
    """字符串转 datetime 格式

    常用格式说明：
        %y: 年（需要补零） 00, 01, …, 99
        %Y: 年 2013, 2014, …, 9999
        %m: 月 01, 02, …, 12
        %d: 日 01, 02, …, 31
        %H: 时 00, 01, …, 23
        %M: 分 00, 01, …, 59
        %S: 秒 00, 01, …, 59

    补零的意思指 '1-2-12' -> '01-02-12'
    官方文档虽然说明都要补零，但实测只有 %y 必须补零才能正常转换

        ref: https://docs.python.org/3/library/datetime.html?highlight=strptime#strftime-and-strptime-behavior

    Args:
        date_str (str):
        fmt (str): 时间格式

    Returns:
        datetime:

    Examples:

        >>> str_to_date('2001-02-12 12:01')
        datetime.datetime(2001, 2, 12, 12, 1)


    """
    # if 'y' in fmt:
    #     try:
    #         date = datetime.strptime(date_string, fmt)
    #     except ValueError:
    #         print('The %y need to pad zero.')
    # else:
    #     date = datetime.strptime(date_string, fmt)

    return datetime.strptime(date_str, fmt)


def date_to_time(date, fmt='%Y-%m-%d %H:%M'):
    """datetime 转字符串

    这个函数配合 str_to_date() 可以用来快速修改时间格式，比如
        '21-01-2008' -> '2008-01-21'

    再比如"补零"：
        '2008-1-2' -> '2008-01-02'

    Args:
        date (datetime):
        fmt (str): 时间格式

    Returns:
        str:

    Examples：

        >>> date = str_to_date('2001-1-12 12:2')
        >>> date_to_time(date, '%Y-%m-%d')
        '2001-01-12'

    """
    return datetime.strftime(date, fmt)


def get_timedelta(time_from, time_to, time_delta,
                  need_str=True, fmt_in='%Y-%m-%d', fmt_out='%Y-%m-%d %H:%M'):
    """获取等间隔的时间序列

    在处理缺失数据时，可以配合 pandas 使用

    Args:
        time_from (str or datetime):
        time_to (str or datetime):
        time_delta (int or timedelta): 如果是 int 默认指分钟
        need_str (bool): 是否需要将时间序列转成字符串
        fmt_in: 输入时间的格式
        fmt_out: 输出时间的格式

    Returns:

    """
    if type(time_from) is str:
        time_from = datetime.strptime(time_from, fmt_in)
        time_to = datetime.strptime(time_to, fmt_in)
    if type(time_delta) is int:
        time_delta = timedelta(minutes=time_delta)

    if need_str:
        return ((time_from + time_delta * i).strftime(fmt_out) for i in range(int((time_to - time_from) / time_delta)))
    else:
        return (time_from + time_delta * i for i in range(int((time_to - time_from) / time_delta)))
